"""
https://github.com/fastai/

For the licence see licence/fastai_licence.txt
"""

from typing import Any
from torch import nn


def _hook_inner(m, i, o):
    return o


def is_listy(x: Any) -> bool:
    return isinstance(x, (tuple, list))


class Hook():
    "Create a hook on `m` with `hook_func`."

    def __init__(self, m: nn.Module, hook_func, detach: bool = True):
        self.hook_func, self.detach, self.stored = hook_func, detach, None
        self.hook = m.register_forward_hook(self.hook_fn)
        self.removed = False

    def hook_fn(self, module: nn.Module, input, output):
        "Applies `hook_func` to `module`, `input`, `output`."
        if self.detach:
            input = (o.detach() for o in input) if is_listy(input) else input.detach()
            output = (o.detach() for o in output) if is_listy(output) else output.detach()
        self.stored = self.hook_func(module, input, output)

    def remove(self):
        if not self.removed:
            self.hook.remove()
            self.removed = True

    def __enter__(self, *args):
        return self

    def __exit__(self, *args):
        self.remove()



from typing import Callable, Dict


class HistoryCondition:
    def __init__(self, metric_name: str, history_condition: Callable, history=()):
        """
        The class aggregates the history of metrics and implements a condition on one of the metrics

        >>> hc = HistoryCondition("accuracy", lambda hist: len(hist) < 2 or hist[-1] > max(hist[:-1]))
        >>> assert hc({"accuracy": 0.10, "precision": 0.5}) is True
        >>> assert hc({"accuracy": 0.08, "precision": 0.6}) is False
        >>> assert hc({"accuracy": 0.09, "precision": 0.7}) is False
        >>> assert hc({"accuracy": 0.11, "precision": 0.8}) is True
        """
        self.metric_name = metric_name
        self.history = list(history[:])
        self.condition = history_condition

    def __call__(self, metrics: Dict):
        self.history.append(metrics[self.metric_name])
        result = self.condition(self.history[:])
        return result



import torch
from pytest import approx
from pytorch_nn_tools.train.metrics.processor import MetricMod

val1 = 10
val2 = 20.1


def test_metric_mod__handles_tensors():
    mm = MetricMod()
    result = mm({
        'metric1': val1,
        'metric2': torch.tensor(val2, requires_grad=True)
    })
    assert result == {
        'metric1': val1,
        'metric2': approx(val2)
    }


def test_metric_mod__handles_names():
    mm = MetricMod(name_fn=lambda x: x + "_new")
    result = mm({
        'metric1': val1,
        'metric2': val2
    })
    assert result == {
        'metric1_new': val1,
        'metric2_new': approx(val2)
    }


def test_processors_aggregation():
    mm1 = MetricMod(name_fn=lambda x: x + "_first")
    mm2 = MetricMod(name_fn=lambda x: x + "_second", value_fn=lambda x: x + 1)
    mm = mm1 + mm2
    result = mm({
        'metric1': val1,
    })
    assert result == {
        'metric1_first_second': val1+1,
    }



from .hook import Hook, _hook_inner

__all__ = ["Hook", "_hook_inner", "get_output_size", "SavingFn", "Hooking"]


def get_output_size(module, input, output):
    return output.size()


class SavingFn:
    def __init__(self, fn):
        self.results = []
        self.fn = fn

    def __call__(self, *args, **kwargs):
        self.results.append(self.fn(*args, **kwargs))


class Hooking:
    def __init__(self,
                 modules_to_hook, hook_fn,
                 register_fn=lambda module, hook: module.register_forward_hook(hook)):
        self.saving = [SavingFn(hook_fn) for _ in modules_to_hook]
        self.handles = [register_fn(module=m, hook=h) for m, h in zip(modules_to_hook, self.saving)]
        self.removed = False

    def remove(self):
        if not self.removed:
            for handle in self.handles:
                handle.remove()
            self.removed = True

    def results(self):
        return [s.results for s in self.saving]

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.remove()



import collections

import torch


def apply_recursively(func, x):
    """
    Apply `func` recursively to `x`.

    >>> result = apply_recursively(\
            lambda val: val+1, {'a': 10, 'b': [20,30,{'c':40}]})
    >>> assert result == {'a': 11, 'b': [21,31,{'c':41}]}
    >>> from collections import OrderedDict
    >>> result = apply_recursively(\
            lambda val: val+1,\
            OrderedDict([ \
                ('a', 10),  \
                ('b', (20,30)) \
            ]) \
        )
    >>> assert result == OrderedDict([('a', 11), ('b', (21,31))])
    >>> assert type(result) == OrderedDict
    >>> assert type(result['b']) == tuple
    """
    if isinstance(x, collections.abc.Sequence):
        return type(x)([apply_recursively(func, i) for i in x])
    if isinstance(x, collections.abc.Mapping):
        return type(x)(
            (k, apply_recursively(func, v))
            for k, v in x.items()
        )
    return func(x)


def to_device(data, device):
    """
    Sends the input to a given device.

    If the input is a Tensor it uses non_blocking=True for speed up.
    Some explanation can be found here https://discuss.pytorch.org/t/should-we-set-non-blocking-to-true/38234/7
    """

    def _inner(x):
        if isinstance(x, torch.Tensor):
            return x.to(device, non_blocking=True)
        elif hasattr(x, "to_device"):
            return x.to_device(device)
        else:
            return x

    return apply_recursively(_inner, data)



import torch
from torch import nn
from typing import Tuple


def dummy_batch(m: nn.Module, size: Tuple) -> torch.Tensor:
    return next(m.parameters()).new_empty(size=size).requires_grad_(False).uniform_(-1., 1.)



import json
import logging
from pathlib import Path
from typing import Union, Optional

import torch


class CheckpointSaver:
    def __init__(self, path: Union[Path, str], logger=None):
        """
        Saves and loads checkpoints of the model together with optimizer and scheduler.
        """
        if logger is None:
            logger = logging.getLogger(__name__)
        self.logger = logger
        self.path = Path(path)
        self.path.mkdir(parents=True, exist_ok=True)

    def save(self, model, optimizer, scheduler, epoch: int):
        """
        Saves the state of the model, optimizer and scheduler to the files corresponding to the given epoch index.
        """
        name_model = self._name_model(epoch)
        self._save_to_fname(model, name_model)

        name_optimizer = self._name_optimizer(epoch)
        self._save_to_fname(optimizer, name_optimizer)

        name_scheduler = self._name_scheduler(epoch)
        self._save_to_fname(scheduler, name_scheduler)

        meta = {
            'epoch': epoch,
            'model': name_model,
            'scheduler': name_scheduler,
            'optimizer': name_optimizer
        }
        path_meta = self.path.joinpath(self._name_meta(epoch))
        with path_meta.open("w") as f:
            json.dump(meta, f)

        self.logger.debug(f"saved meta data {meta} to {path_meta}")

    def load(self, model, optimizer, scheduler, epoch: int):
        """
        Loads the state of the model, optimizer and scheduler from the files corresponding to the given epoch index.
        """
        path_meta = self.path.joinpath(self._name_meta(epoch))
        with path_meta.open() as f:
            meta = json.load(f)

        name_model = meta['model']
        name_optimizer = meta['optimizer']
        name_scheduler = meta['scheduler']

        self._load_module_with_update('model', model, self.path.joinpath(name_model))
        self._load_module_optional('optimizer', optimizer, self.path.joinpath(name_optimizer))
        self._load_module_optional('scheduler', scheduler, self.path.joinpath(name_scheduler))

    def find_last(self, start_epoch, end_epoch) -> Optional[int]:
        for epoch in range(end_epoch, start_epoch - 1, -1):
            path = self.path.joinpath(self._name_meta(epoch))
            if path.exists():
                return epoch
        return None

    def _save_to_fname(self, module, fname: str):
        torch.save(module.state_dict(), self.path.joinpath(fname))

    def _name_meta(self, epoch: int) -> str:
        return "epoch_{epoch:05d}.meta.json".format(epoch=epoch)

    def _name_scheduler(self, epoch: int) -> str:
        return "epoch_{epoch:05d}.scheduler.pth".format(epoch=epoch)

    def _name_optimizer(self, epoch: int) -> str:
        return "epoch_{epoch:05d}.optimizer.pth".format(epoch=epoch)

    def _name_model(self, epoch: int) -> str:
        return "epoch_{epoch:05d}.pth".format(epoch=epoch)

    def _load_module_with_update(self, module_name: str, dst_module, path: Path):
        self.logger.debug(f"loading {module_name} from {path}")
        pretrained_dict = torch.load(path)
        module_dict = dst_module.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in module_dict}
        module_dict.update(pretrained_dict)
        dst_module.load_state_dict(module_dict)

    def _load_module_optional(self, module_name: str, dst_module, path: Path):
        if path.exists():
            self.logger.debug(f"loading {module_name} state from {path}")
            module_dict = torch.load(path)
            dst_module.load_state_dict(module_dict)
        else:
            self.logger.debug(f"{module_name} state not found")



from typing import List

import torch
from dataclasses import dataclass
from torch import nn


class NameGetter:
    def __init__(self, unnamed_template='unnamed_{count}_{type_name}'):
        self.count = 0
        try:
            unnamed_template.format(count=self.count, type_name="dummy2")
        except KeyError as e:
            raise ValueError("Invalid template '{}'. There is an issue with key: {}".format(unnamed_template, e))
        self.unnamed_template = unnamed_template

    def __call__(self, module):
        try:
            name = module.name
        except AttributeError:
            name = self.unnamed_template.format(count=self.count, type_name=type(module).__name__)
            self.count += 1
        return name


@dataclass
class LayerSize:
    name: str
    module: nn.Module
    size: torch.Size


def get_output_sizes(model, x) -> List[LayerSize]:
    sizes = []
    hooks = []
    get_name = NameGetter()

    def hook(module, input, output):
        name = get_name(module)
        sizes.append(LayerSize(name, module, output.size()))

    def register_hook(module):
        hooks.append(module.register_forward_hook(hook))

    # register hook
    model.apply(register_hook)

    # make a forward pass to trigger the hooks
    with torch.no_grad():
        model(x)

    # remove these hooks
    for h in hooks:
        h.remove()

    return sizes



"""Top-level package for pytorch-nn-tools."""

__author__ = """SergeM"""
__email__ = 'serge-m@users.noreply.github.com'

from .version import __version__

from .hook import Hook, _hook_inner
from .structure import get_output_sizes, LayerSize, NameGetter
from .memory import dummy_batch


__all__ = ["Hook", "_hook_inner", "get_output_sizes", "LayerSize", "NameGetter", "dummy_batch", "__version__"]



from typing import Any

import pytest
import torch
from pytorch_nn_tools.hook import Hooking, get_output_size
from torch import nn


class DummyLinear(nn.Module):
    """
    Dummy linear class with predefined weights for testing.
    """

    def __init__(self, weight, bias):
        super().__init__()
        assert weight.shape[0] == bias.shape[0]
        assert weight.ndim == 2
        assert bias.ndim == 1
        self.linear = nn.Linear(weight.shape[1], weight.shape[0])
        with torch.no_grad():
            self.linear.weight.copy_(weight)
            self.linear.bias.copy_(bias)

    def forward(self, *input: Any, **kwargs: Any):
        return self.linear(*input, **kwargs)


@pytest.fixture
def linear1():
    return DummyLinear(weight=torch.tensor([[1., 0.], [0., 1.]]), bias=torch.tensor([0.6, 0.8]))


@pytest.fixture
def linear2():
    return DummyLinear(weight=torch.tensor([[100, 1]]), bias=torch.tensor([0.09]))


@pytest.fixture
def x():
    return torch.tensor((5., 7.))


def test_dummy_linear(linear1, linear2, x):
    with torch.no_grad():
        r1 = linear1(x)
        r2 = linear2(r1)

    assert r1.allclose(torch.tensor((5.6, 7.8)))
    assert r2.allclose(torch.tensor(567.89))


def test_hooking_and_get_output_size(linear1, linear2, x):
    module = nn.Sequential(linear1, linear2)
    with Hooking([linear1, linear2], get_output_size) as h:
        module(x)

    assert h.results() == [[torch.Size([2])], [torch.Size([1])]]


def test_hooking_output_results(linear1, linear2, x):
    module = nn.Sequential(linear1, linear2)

    with Hooking([linear1], lambda module_, input, output: output.detach()) as h:
        module(x)
        module(x * 0)

    assert len(h.results()) == 1
    assert len(h.results()[0]) == 2
    for r, expected in zip(
        h.results()[0],
        [torch.tensor([5.6, 7.8]), torch.tensor([0.6, 0.8])]
    ):
        assert r.allclose(expected)




from typing import Callable, Iterable, Dict


def map_dict(dictionary,
             key_fn: Callable = lambda key: key,
             value_fn: Callable = lambda value: value,
             skip_keys: Iterable = ()) -> Dict:
    """
    Converts all the keys and values of a dictionary according to the
    specified mapping functions, except for the (key, value) pairs where
    `key` belongs to `skip_keys`.

    >>> data = {'a': 1, 'b': 2, 'c':3}
    >>> expected = {'A': 11, 'b': 2, 'C': 13}
    >>> assert map_dict(data, lambda key: key.upper(), lambda value: value + 10, skip_keys='b') == expected

    >>> from collections import OrderedDict
    >>> data = OrderedDict([(i, i+10) for i in range(1000, -1, -1)] )
    >>> expected = dict([(i+1, i+12) for i in range(1000, -1, -1)] )
    >>> assert map_dict(data, lambda key: key + 1 , lambda value: value + 2).items() == expected.items()

    """
    skip_keys = set(skip_keys)
    return type(dictionary)([
        (
            (key_fn(key), value_fn(value))
            if key not in skip_keys
            else (key, value)
        )
        for key, value in dictionary.items()
    ])



from collections import defaultdict
from pathlib import Path
from typing import Dict, Union, Any, Iterable
import enum

import torch
from pytorch_nn_tools.convert import map_dict
from torch.utils.tensorboard import SummaryWriter


@enum.unique
class Marker(enum.Enum):
    ITERATION = 'iteration'
    EPOCH = 'epoch'


MetricNameType = Union[str, Marker]
MetricType = Dict[MetricNameType, Any]


class MetricProcessor:
    def __call__(self, data: MetricType) -> MetricType:
        return data

    def __add__(self, other):
        processors = []
        for x in [self, other]:
            if isinstance(x, MetricPipeline):
                processors.extend(x._processors)
            elif isinstance(x, MetricProcessor):
                processors.append(x)
            else:
                raise ValueError(f"Cannot add MetricProcessor and {other} of type {type(other)}")
        return MetricPipeline(*processors)


class MetricPipeline(MetricProcessor):
    def __init__(self, *processors):
        self._processors = processors

    def __call__(self, data: MetricType) -> MetricType:
        for p in self._processors:
            data = p(data)
        return data


class MetricAggregator(MetricProcessor):
    DEFAULT_SKIPPED = tuple(Marker)

    def __init__(self, skip_names: Iterable[MetricNameType] = DEFAULT_SKIPPED):
        """
        Aggregates metrics values over time
        >>> agg = MetricAggregator(["b"])
        >>> input1 = {'a': 1, 'b': 10}
        >>> input2 = {'a': 5, 'b': 20}
        >>> ret = agg(input1)
        >>> assert ret == input1
        >>> ret = agg(input2)
        >>> assert ret == input2
        >>> assert agg.aggregate() == {'a': (1+5)/2}
        """
        self._metrics = defaultdict(list)
        self.skipped = set(skip_names)

    def __call__(self, data: MetricType) -> MetricType:
        for k, v in data.items():
            if k not in self.skipped:
                self._metrics[k].append(v)
        return data

    def aggregate(self) -> Dict:
        return {
            k: sum(vs) / (len(vs) if vs else 1.)
            for k, vs in self._metrics.items()
        }


def detach_and_get_if_tensor(x):
    return x.detach().item() if torch.is_tensor(x) else x


class MetricMod(MetricProcessor):
    DEFAULT_SKIPPED = tuple(Marker)

    def __init__(self, name_fn=lambda x: x, value_fn=detach_and_get_if_tensor,
                 skip_names: Iterable[MetricNameType] = DEFAULT_SKIPPED):
        self.name_fn = name_fn
        self.value_fn = value_fn
        self.skipped = set(skip_names)

    def __call__(self, data: MetricType) -> MetricType:
        return map_dict(data, key_fn=self.name_fn, value_fn=self.value_fn, skip_keys=self.skipped)


class MetricLogger(MetricProcessor):
    def __init__(self, path: Union[str, Path]):
        path = Path(path)
        path.mkdir(exist_ok=True, parents=True)
        self.writer = SummaryWriter(path)

    def __call__(self, data: MetricType) -> MetricType:
        iteration = data.pop(Marker.ITERATION, None)
        if iteration is None:
            iteration = data.pop(Marker.EPOCH, None)
        if iteration is not None:
            for name, value in data.items():
                self.writer.add_scalar(name, value, iteration)

        return data

    def close(self):
        self.writer.close()


mod_name_train = MetricMod(
    name_fn=lambda name: f"train.{name}",
)
mod_name_val = MetricMod(
    name_fn=lambda name: f"val.{name}",
)



class ProgressTracker:
    def __init__(self):
        """
        Keeps track of global number of iterations.
        >>> tracker = ProgressTracker()
        >>> for i in tracker.track((1,2,3)): assert tracker.cnt_total_iter == i
        >>> assert tracker.cnt_total_iter == 3
        >>> for i in tracker.track((1,2,3,4)): assert tracker.cnt_total_iter == i + 3
        >>> assert tracker.cnt_total_iter == 7
        """
        self.cnt_total_iter = 0

    def track(self, dl):
        def tracked_iterator():
            for x in dl:
                self.cnt_total_iter += 1
                yield x

        return tracked_iterator()



from collections import OrderedDict

from pytorch_nn_tools.convert import map_dict


def test_convert_keeps_type():
    data = {'apple': 4, 'banana': 3}
    result = map_dict(data, lambda key: key, lambda value: value+1)
    assert type(result) == dict

    data = OrderedDict({'apple': 4, 'banana': 3})
    result = map_dict(data, lambda key: key, lambda value: value+1)
    assert type(result) == OrderedDict
    assert result == OrderedDict({'apple': 5, 'banana': 4})



import PIL.Image
import numpy as np
import torch
import torchvision.transforms.functional as tvf

imagenet_stats = dict(mean=[0.485, 0.456, 0.406],
                      std=[0.229, 0.224, 0.225])


class UnNormalize_(object):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std

    def __call__(self, tensor):
        for t, m, s in zip(tensor, self.mean, self.std):
            t.mul_(s).add_(m)
        return tensor


def tfm_vis_img(tensor, size=None, unnormalize_img=UnNormalize_(**imagenet_stats)):
    unnormalized = unnormalize_img(tensor.detach().clone())
    img = tvf.to_pil_image(unnormalized, mode='RGB')
    if size is not None:
        img = tvf.resize(img, size, interpolation=PIL.Image.NEAREST)
    return np.array(img)


def tfm_vis_mask(tensor, size=None):
    img = tvf.to_pil_image(tensor.detach().type(torch.IntTensor), mode='I')
    if size is not None:
        img = tvf.resize(img, size, interpolation=PIL.Image.NEAREST)
    return np.array(img)


DEFAULT_KWARGS_IMG = {'interpolation': 'nearest'}
DEFAULT_KWARGS_MASK = {'interpolation': 'nearest', 'cmap': 'tab20', 'vmin': 0, 'vmax': 20}


class ImgShow:
    def __init__(self, ax=None, size=None,
                 tfm_img=tfm_vis_img, tfm_mask=tfm_vis_mask,
                 show_kwargs_img=None, show_kwargs_mask=None):
        """
        Class for visualization of tensors representing images.
        Sample usage:
        >>> from pytorch_nn_tools.visual import ImgShow
        >>> import matplotlib.pyplot as plt # doctest: +SKIP
        >>> ish = ImgShow(ax=plt) # doctest: +SKIP
        >>> _ = ish.show_image(torch.rand(3, 10, 20)) # doctest: +SKIP
        """
        if show_kwargs_mask is None:
            show_kwargs_mask = DEFAULT_KWARGS_MASK
        if show_kwargs_img is None:
            show_kwargs_img = DEFAULT_KWARGS_IMG
        self.ax = ax
        self.size = size
        self.tfm_img = tfm_img
        self.tfm_mask = tfm_mask
        self.show_kwargs_img = show_kwargs_img
        self.show_kwargs_mask = show_kwargs_mask

    def with_axes(self, ax):
        return ImgShow(ax=ax, size=self.size, tfm_img=self.tfm_img, tfm_mask=self.tfm_mask,
                       show_kwargs_img=self.show_kwargs_img,
                       show_kwargs_mask=self.show_kwargs_mask)

    def with_size(self, size):
        return ImgShow(ax=self.ax, size=size, tfm_img=self.tfm_img, tfm_mask=self.tfm_mask,
                       show_kwargs_img=self.show_kwargs_img,
                       show_kwargs_mask=self.show_kwargs_mask
                       )

    def show_image(self, tensor):
        self._check_axes()
        img = self.tfm_img(tensor, size=self.size)
        self.ax.imshow(img, **self.show_kwargs_img)
        return self

    def show_mask(self, tensor):
        self._check_axes()
        img = self.tfm_mask(tensor, size=self.size)
        self.ax.imshow(img, **self.show_kwargs_mask)
        return self

    def _check_axes(self):
        if self.ax is None:
            raise ValueError("Axes are not initialized for ImageShow object")
